@@ -1,5 +1,5 @@
 /* libcap-ng.c --
- * Copyright 2009-10, 2013, 2017 Red Hat Inc., Durham, North Carolina.
+ * Copyright 2009-10, 2013, 2017, 2020 Red Hat Inc.
  * All Rights Reserved.
  *
  * This library is free software; you can redistribute it and/or
@@ -33,6 +33,7 @@
 #include <sys/stat.h>
 #include <errno.h>
 #include <fcntl.h>
+#include <endian.h>
 #include <byteswap.h>
 #ifdef HAVE_PTHREAD_H
 #include <pthread.h>	// For pthread_atfork
@@ -51,12 +52,20 @@ int last_cap hidden = -1;
  * 2.6.24 kernel	XATTR_NAME_CAPS
  * 2.6.25 kernel	PR_CAPBSET_DROP, CAPABILITY_VERSION_2
  * 2.6.26 kernel	PR_SET_SECUREBITS, SECURE_*_LOCKED, VERSION_3
+ * 4.3    kernel	PR_CAP_AMBIENT
+ * 4.14   kernel	VFS_CAP_REVISION_3
  */
 
 /* External syscall prototypes */
 extern int capset(cap_user_header_t header, cap_user_data_t data);
 extern int capget(cap_user_header_t header, const cap_user_data_t data);
 
+// Local functions
+static void update_bounding_set(capng_act_t action, unsigned int capability,
+	unsigned int idx);
+static void update_ambient_set(capng_act_t action, unsigned int capability,
+	unsigned int idx);
+
 // Local defines
 #define MASK(x) (1U << (x))
 #ifdef PR_CAPBSET_DROP
@@ -68,7 +77,7 @@ extern int capget(cap_user_header_t header, const cap_user_data_t data);
 
 // Re-define cap_valid so its uniform between V1 and V3
 #undef cap_valid
-#define cap_valid(x) ((x) <= last_cap)
+#define cap_valid(x) ((x) <= (unsigned int)last_cap)
 
 // If we don't have the xattr library, then we can't
 // compile-in file system capabilities
@@ -122,12 +131,21 @@ extern int capget(cap_user_header_t header, const cap_user_data_t data);
 #endif
 /* Setuid apps run by uid 0 don't get caps back */
 #ifndef SECURE_NO_SETUID_FIXUP
-#define SECURE_NO_SETUID_FIXUP          2  
+#define SECURE_NO_SETUID_FIXUP          2
 #endif
 #ifndef SECURE_NO_SETUID_FIXUP_LOCKED
 #define SECURE_NO_SETUID_FIXUP_LOCKED   3  /* make bit-2 immutable */
 #endif
 
+#ifndef VFS_CAP_U32
+#define VFS_CAP_U32 2
+#endif
+
+#if (VFS_CAP_U32 != 2)
+#error VFS_CAP_U32 does not match the library, you need a new version
+#endif
+
+
 // States: new, allocated, initted, updated, applied
 typedef enum { CAPNG_NEW, CAPNG_ERROR, CAPNG_ALLOCATED, CAPNG_INIT,
 	CAPNG_UPDATED, CAPNG_APPLIED } capng_states_t;
@@ -135,24 +153,28 @@ typedef enum { CAPNG_NEW, CAPNG_ERROR, CAPNG_ALLOCATED, CAPNG_INIT,
 // Create an easy data struct out of the kernel definitions
 typedef union {
 	struct __user_cap_data_struct v1;
-	struct __user_cap_data_struct v3[2];
+	struct __user_cap_data_struct v3[VFS_CAP_U32];
 } cap_data_t;
 
 // This struct keeps all state info
 struct cap_ng
 {
 	int cap_ver;
+	int vfs_cap_ver;
 	struct __user_cap_header_struct hdr;
 	cap_data_t data;
 	capng_states_t state;
-	__u32 bounds[2];
+	__le32 rootid;
+	__u32 bounds[VFS_CAP_U32];
+	__u32 ambient[VFS_CAP_U32];
 };
 
 // Global variables with per thread uniqueness
-static __thread struct cap_ng m =	{ 1,
+static __thread struct cap_ng m =	{ 1, 1,
 					{0, 0},
 					{ {0, 0, 0} },
-					CAPNG_NEW,
+					CAPNG_NEW, CAPNG_UNSET_ROOTID,
+					{0, 0},
 					{0, 0} };
 
 /*
@@ -166,7 +188,9 @@ static void deinit(void)
 static void init_lib(void) __attribute__ ((constructor));
 static void init_lib(void)
 {
+#ifdef HAVE_PTHREAD_H
 	pthread_atfork(NULL, NULL, deinit);
+#endif
 }
 
 static void init(void)
@@ -186,6 +210,12 @@ static void init(void)
 		return;
 	}
 
+#if VFS_CAP_REVISION == VFS_CAP_REVISION_1
+	m.vfs_cap_ver = 1;
+#else
+	m.vfs_cap_ver = 2; // Intentionally set to 2 for both 2 & 3
+#endif
+
 	memset(&m.data, 0, sizeof(cap_data_t));
 #ifdef HAVE_SYSCALL_H
 	m.hdr.pid = (unsigned)syscall(__NR_gettid);
@@ -217,6 +247,7 @@ static void init(void)
 		if (last_cap == -1)
 			last_cap = CAP_LAST_CAP;
 	}
+	m.rootid = CAPNG_UNSET_ROOTID;
 	m.state = CAPNG_ALLOCATED;
 }
 
@@ -232,6 +263,10 @@ void capng_clear(capng_select_t set)
 #ifdef PR_CAPBSET_DROP
 	if (set & CAPNG_SELECT_BOUNDS)
 		memset(m.bounds, 0, sizeof(m.bounds));
+#endif
+#ifdef PR_CAP_AMBIENT
+	if (set & CAPNG_SELECT_AMBIENT)
+		memset(m.ambient, 0, sizeof(m.ambient));
 #endif
 	m.state = CAPNG_INIT;
 }
@@ -263,6 +298,13 @@ void capng_fill(capng_select_t set)
 		for (i=0; i<sizeof(m.bounds)/sizeof(__u32); i++)
 			m.bounds[i] = 0xFFFFFFFFU;
 	}
+#endif
+#ifdef PR_CAP_AMBIENT
+	if (set & CAPNG_SELECT_AMBIENT) {
+		unsigned i;
+		for (i=0; i<sizeof(m.ambient)/sizeof(__u32); i++)
+			m.ambient[i] = 0xFFFFFFFFU;
+	}
 #endif
 	m.state = CAPNG_INIT;
 }
@@ -277,11 +319,41 @@ void capng_setpid(int pid)
 	m.hdr.pid = pid;
 }
 
+int capng_get_rootid(void)
+{
+#ifdef VFS_CAP_REVISION_3
+	return m.rootid;
+#else
+	return CAPNG_UNSET_ROOTID;
+#endif
+}
+
+int capng_set_rootid(int rootid)
+{
+#ifdef VFS_CAP_REVISION_3
+	if (m.state == CAPNG_NEW)
+		init();
+	if (m.state == CAPNG_ERROR)
+		return -1;
+
+	if (rootid < 0)
+		return -1;
+
+	m.rootid = rootid;
+	m.vfs_cap_ver = 3;
+
+	return 0;
+#else
+	return -1;
+#endif
+}
+
 #ifdef PR_CAPBSET_DROP
 static int get_bounding_set(void)
 {
 	char buf[64];
 	FILE *f;
+	int rc;
 
 	snprintf(buf, sizeof(buf), "/proc/%d/status", m.hdr.pid ? m.hdr.pid :
 #ifdef HAVE_SYSCALL_H
@@ -290,18 +362,79 @@ static int get_bounding_set(void)
 		(int)getpid();
 #endif
 	f = fopen(buf, "re");
-	if (f == NULL)
-		return -1;
-	__fsetlocking(f, FSETLOCKING_BYCALLER);
-	while (fgets(buf, sizeof(buf), f)) {
-		if (strncmp(buf, "CapB", 4))
-			continue;
-		sscanf(buf, "CapBnd:  %08x%08x", &m.bounds[1], &m.bounds[0]);
+	if (f) {
+		__fsetlocking(f, FSETLOCKING_BYCALLER);
+		while (fgets(buf, sizeof(buf), f)) {
+			if (strncmp(buf, "CapB", 4))
+				continue;
+			sscanf(buf, "CapBnd:  %08x%08x",
+			       &m.bounds[1], &m.bounds[0]);
+			fclose(f);
+			return 0;
+		}
 		fclose(f);
-		return 0;
+		return -1;
 	}
-	fclose(f);
-	return -1;
+	// Might be in a container with no procfs - do it the hard way
+	memset(m.bounds, 0, sizeof(m.bounds));
+	unsigned int i = 0;
+	do {
+		rc = prctl(PR_CAPBSET_READ, i, 0, 0, 0);
+		if (rc < 0)
+			return -1;
+
+		// Just add set bits
+		if (rc)
+			update_bounding_set(CAPNG_ADD, i%32, i>>5);
+		i++;
+	} while (cap_valid(i));
+
+	return 0;
+}
+#endif
+
+#ifdef PR_CAP_AMBIENT
+static int get_ambient_set(void)
+{
+	char buf[64];
+	FILE *f;
+	int rc;
+
+	snprintf(buf, sizeof(buf), "/proc/%d/status", m.hdr.pid ? m.hdr.pid :
+#ifdef HAVE_SYSCALL_H
+		(int)syscall(__NR_gettid));
+#else
+		(int)getpid();
+#endif
+	f = fopen(buf, "re");
+	if (f) {
+		__fsetlocking(f, FSETLOCKING_BYCALLER);
+		while (fgets(buf, sizeof(buf), f)) {
+			if (strncmp(buf, "CapA", 4))
+				continue;
+			sscanf(buf, "CapAmb:  %08x%08x",
+			       &m.ambient[1], &m.ambient[0]);
+			fclose(f);
+			return 0;
+		}
+		fclose(f);
+		return -1;
+	}
+	// Might be in a container with no procfs - do it the hard way
+	memset(m.ambient, 0, sizeof(m.ambient));
+	unsigned int i = 0;
+	do {
+		rc = prctl(PR_CAP_AMBIENT, PR_CAP_AMBIENT_IS_SET, i, 0, 0);
+		if (rc < 0)
+			return -1;
+
+		// Just add set bits
+		if (rc)
+			update_ambient_set(CAPNG_ADD, i%32, i>>5);
+		i++;
+	} while (cap_valid(i));
+
+	return 0;
 }
 #endif
 
@@ -321,6 +454,11 @@ int capng_get_caps_process(void)
 		rc = get_bounding_set();
 		if (rc < 0)
 			m.state = CAPNG_ERROR;
+#endif
+#ifdef PR_CAP_AMBIENT
+		rc = get_ambient_set();
+		if (rc < 0)
+			m.state = CAPNG_ERROR;
 #endif
 	}
 
@@ -328,10 +466,14 @@ int capng_get_caps_process(void)
 }
 
 #ifdef VFS_CAP_U32
+#ifdef VFS_CAP_REVISION_3
+static int load_data(const struct vfs_ns_cap_data *filedata, int size)
+#else
 static int load_data(const struct vfs_cap_data *filedata, int size)
+#endif
 {
 	unsigned int magic;
-	
+
 	if (m.cap_ver == 1)
 		return -1;	// Should never get here but just in case
 
@@ -339,15 +481,22 @@ static int load_data(const struct vfs_cap_data *filedata, int size)
 	switch (magic & VFS_CAP_REVISION_MASK)
 	{
 		case VFS_CAP_REVISION_1:
-			m.cap_ver = 1;
+			m.vfs_cap_ver = 1;
 			if (size != XATTR_CAPS_SZ_1)
 				return -1;
 			break;
 		case VFS_CAP_REVISION_2:
-			m.cap_ver = 2;
+			m.vfs_cap_ver = 2;
 			if (size != XATTR_CAPS_SZ_2)
 				return -1;
 			break;
+#ifdef VFS_CAP_REVISION_3
+		case VFS_CAP_REVISION_3:
+			m.vfs_cap_ver = 3;
+			if (size != XATTR_CAPS_SZ_3)
+				return -1;
+			break;
+#endif
 		default:
 			return -1;
 	}
@@ -366,6 +515,12 @@ static int load_data(const struct vfs_cap_data *filedata, int size)
 		m.data.v3[0].effective = 0;
 		m.data.v3[1].effective = 0;
 	}
+#ifdef VFS_CAP_REVISION_3
+	if (size == XATTR_CAPS_SZ_3) {
+	    struct vfs_ns_cap_data *d = (struct vfs_ns_cap_data *)filedata;
+	    m.rootid = FIXUP(d->rootid);
+	}
+#endif
 	return 0;
 }
 #endif
@@ -376,8 +531,11 @@ int capng_get_caps_fd(int fd)
 	return -1;
 #else
 	int rc;
+#ifdef VFS_CAP_REVISION_3
+	struct vfs_ns_cap_data filedata;
+#else
 	struct vfs_cap_data filedata;
-
+#endif
 	if (m.state == CAPNG_NEW)
 		init();
 	if (m.state == CAPNG_ERROR)
@@ -441,6 +599,17 @@ static void update_bounding_set(capng_act_t action, unsigned int capability,
 #endif
 }
 
+static void update_ambient_set(capng_act_t action, unsigned int capability,
+	unsigned int idx)
+{
+#ifdef PR_CAP_AMBIENT
+	if (action == CAPNG_ADD)
+		m.ambient[idx] |= MASK(capability);
+	else
+		m.ambient[idx] &= ~(MASK(capability));
+#endif
+}
+
 int capng_update(capng_act_t action, capng_type_t type, unsigned int capability)
 {
 	// Before updating, we expect that the data is initialized to something
@@ -459,7 +628,7 @@ int capng_update(capng_act_t action, capng_type_t type, unsigned int capability)
 		if (CAPNG_INHERITABLE & type)
 			v1_update(action, capability, &m.data.v1.inheritable);
 	} else {
-		int idx;
+		unsigned int idx;
 
 		if (capability > 31) {
 			idx = capability>>5;
@@ -475,6 +644,8 @@ int capng_update(capng_act_t action, capng_type_t type, unsigned int capability)
 			update_inheritable(action, capability, idx);
 		if (CAPNG_BOUNDING_SET & type)
 			update_bounding_set(action, capability, idx);
+		if (CAPNG_AMBIENT & type)
+			update_ambient_set(action, capability, idx);
 	}
 
 	m.state = CAPNG_UPDATED;
@@ -520,13 +691,14 @@ int capng_apply(capng_select_t set)
 	if (m.state < CAPNG_INIT)
 		return -1;
 
-	if (set & CAPNG_SELECT_BOUNDS) { 
+	if (set & CAPNG_SELECT_BOUNDS) {
 #ifdef PR_CAPBSET_DROP
-		void *s = capng_save_state();
+		struct cap_ng state;
+		memcpy(&state, &m, sizeof(state)); /* save state */
 		capng_get_caps_process();
 		if (capng_have_capability(CAPNG_EFFECTIVE, CAP_SETPCAP)) {
-			int i;
-			capng_restore_state(&s);
+			unsigned int i;
+			memcpy(&m, &state, sizeof(m)); /* restore state */
 			rc = 0;
 			for (i=0; i <= last_cap && rc == 0; i++)
 				if (capng_have_capability(CAPNG_BOUNDING_SET,
@@ -535,7 +707,30 @@ int capng_apply(capng_select_t set)
 			if (rc == 0)
 				m.state = CAPNG_APPLIED;
 		} else
-			capng_restore_state(&s);
+			memcpy(&m, &state, sizeof(m)); /* restore state */
+#else
+		rc = 0;
+#endif
+	}
+	if (set & CAPNG_SELECT_AMBIENT) {
+#ifdef PR_CAP_AMBIENT
+		unsigned int i;
+		if (capng_have_capabilities(CAPNG_SELECT_AMBIENT) ==
+								CAPNG_NONE) {
+			rc = prctl(PR_CAP_AMBIENT,
+					   PR_CAP_AMBIENT_CLEAR_ALL, 0, 0, 0);
+		} else {
+			// Clear them all
+			rc = prctl(PR_CAP_AMBIENT,
+				   PR_CAP_AMBIENT_CLEAR_ALL, 0, 0, 0);
+			for (i=0; i <= last_cap && rc == 0; i++) {
+				if (capng_have_capability(CAPNG_AMBIENT, i))
+					rc = prctl(PR_CAP_AMBIENT,
+						PR_CAP_AMBIENT_RAISE, i, 0, 0);
+			}
+		}
+		if (rc == 0)
+			m.state = CAPNG_APPLIED;
 #else
 		rc = 0;
 #endif
@@ -550,15 +745,19 @@ int capng_apply(capng_select_t set)
 }
 
 #ifdef VFS_CAP_U32
+#ifdef VFS_CAP_REVISION_3
+static int save_data(struct vfs_ns_cap_data *filedata, int *size)
+#else
 static int save_data(struct vfs_cap_data *filedata, int *size)
+#endif
 {
 	// Now stuff the data structures
-	if (m.cap_ver == 1) {
+	if (m.vfs_cap_ver == 1) {
 		filedata->data[0].permitted = FIXUP(m.data.v1.permitted);
 		filedata->data[0].inheritable = FIXUP(m.data.v1.inheritable);
 		filedata->magic_etc = FIXUP(VFS_CAP_REVISION_1);
 		*size = XATTR_CAPS_SZ_1;
-	} else {
+	} else if (m.vfs_cap_ver == 2 || m.vfs_cap_ver == 3) {
 		int eff;
 
 		if (m.data.v3[0].effective || m.data.v3[1].effective)
@@ -572,6 +771,15 @@ static int save_data(struct vfs_cap_data *filedata, int *size)
 		filedata->magic_etc = FIXUP(VFS_CAP_REVISION_2 | eff);
 		*size = XATTR_CAPS_SZ_2;
 	}
+#ifdef VFS_CAP_REVISION_3
+	if (m.vfs_cap_ver == 3) {
+		// Kernel doesn't support namespaces with non-0 rootid
+		if (m.rootid!= 0)
+			return -1;
+		filedata->rootid = FIXUP(m.rootid);
+		*size = XATTR_CAPS_SZ_3;
+	}
+#endif
 
 	return 0;
 }
@@ -582,8 +790,12 @@ int capng_apply_caps_fd(int fd)
 #ifndef VFS_CAP_U32
 	return -1;
 #else
-	int rc, size;
+	int rc, size = 0;
+#ifdef VFS_CAP_REVISION_3
+	struct vfs_ns_cap_data filedata;
+#else
 	struct vfs_cap_data filedata;
+#endif
 	struct stat buf;
 
 	// Before updating, we expect that the data is initialized to something
@@ -599,7 +811,11 @@ int capng_apply_caps_fd(int fd)
 	if (capng_have_capabilities(CAPNG_SELECT_CAPS) == CAPNG_NONE)
 		rc = fremovexattr(fd, XATTR_NAME_CAPS);
 	else {
-		save_data(&filedata, &size);
+		if (save_data(&filedata, &size)) {
+			m.state = CAPNG_ERROR;
+			errno = EINVAL;
+			return -2;
+		}
 		rc = fsetxattr(fd, XATTR_NAME_CAPS, &filedata, size, 0);
 	}
 
@@ -610,7 +826,7 @@ int capng_apply_caps_fd(int fd)
 #endif
 }
 
-// Change uids keeping/removing only certain capabilities 
+// Change uids keeping/removing only certain capabilities
 // flag to drop supp groups
 int capng_change_id(int uid, int gid, capng_flags_t flag)
 {
@@ -623,7 +839,7 @@ int capng_change_id(int uid, int gid, capng_flags_t flag)
 	// Check the current capabilities
 #ifdef PR_CAPBSET_DROP
 	// If newer kernel, we need setpcap to change the bounding set
-	if (capng_have_capability(CAPNG_EFFECTIVE, CAP_SETPCAP) == 0 && 
+	if (capng_have_capability(CAPNG_EFFECTIVE, CAP_SETPCAP) == 0 &&
 					flag & CAPNG_CLEAR_BOUNDING)
 		capng_update(CAPNG_ADD,
 				CAPNG_EFFECTIVE|CAPNG_PERMITTED, CAP_SETPCAP);
@@ -652,9 +868,13 @@ int capng_change_id(int uid, int gid, capng_flags_t flag)
 	if (rc < 0)
 		return -3;
 
+	// If we are clearing ambient, only clear since its applied at the end
+	if (flag & CAPNG_CLEAR_AMBIENT)
+		capng_clear(CAPNG_SELECT_AMBIENT);
+
 	// Clear bounding set if needed while we have CAP_SETPCAP
 	if (flag & CAPNG_CLEAR_BOUNDING) {
-		capng_clear(CAPNG_BOUNDING_SET);
+		capng_clear(CAPNG_SELECT_BOUNDS);
 		rc = capng_apply(CAPNG_SELECT_BOUNDS);
 		if (rc)
 			return -8;
@@ -708,7 +928,7 @@ int capng_change_id(int uid, int gid, capng_flags_t flag)
 	// Now drop setpcap & apply
 	capng_update(CAPNG_DROP, CAPNG_EFFECTIVE|CAPNG_PERMITTED,
 				CAP_SETPCAP);
-	rc = capng_apply(CAPNG_SELECT_CAPS);
+	rc = capng_apply(CAPNG_SELECT_CAPS|CAPNG_SELECT_AMBIENT);
 	if (rc < 0)
 		return -9;
 
@@ -793,13 +1013,28 @@ capng_results_t capng_have_capabilities(capng_select_t set)
 		else
 			return CAPNG_PARTIAL;
 	}
-#endif 
-
+#endif
+#ifdef PR_CAP_AMBIENT
+	if (set & CAPNG_SELECT_AMBIENT) {
+		if (m.ambient[0] == 0)
+			empty = 1;
+		else if (m.ambient[0] == 0xFFFFFFFFU)
+			full = 1;
+		else
+			return CAPNG_PARTIAL;
+		if ((m.ambient[1] & UPPER_MASK) == 0)
+			empty = 1;
+		else if ((m.ambient[1] & UPPER_MASK) == UPPER_MASK)
+			full = 1;
+		else
+			return CAPNG_PARTIAL;
+	}
+#endif
 	if (empty == 1 && full == 0)
 		return CAPNG_NONE;
 	else if (empty == 0 && full == 1)
 		return CAPNG_FULL;
-	
+
 	return CAPNG_PARTIAL;
 }
 
@@ -822,6 +1057,7 @@ capng_results_t capng_have_permitted_capabilities(void)
 		full = 1;
 	else
 		return CAPNG_PARTIAL;
+
 	if ((m.data.v3[1].permitted & UPPER_MASK) == 0 && !full)
 		empty = 1;
 	else if ((m.data.v3[1].permitted & UPPER_MASK) == UPPER_MASK && !empty)
@@ -833,7 +1069,7 @@ capng_results_t capng_have_permitted_capabilities(void)
 		return CAPNG_NONE;
 	else if (empty == 0 && full == 1)
 		return CAPNG_FULL;
-	
+
 	return CAPNG_PARTIAL;
 }
 
@@ -861,6 +1097,15 @@ static int bounds_bit_check(unsigned int capability, unsigned int idx)
 #endif
 }
 
+static int ambient_bit_check(unsigned int capability, unsigned int idx)
+{
+#ifdef PR_CAP_AMBIENT
+	return MASK(capability) & m.ambient[idx] ? 1 : 0;
+#else
+	return 0;
+#endif
+}
+
 static int v1_check(unsigned int capability, __u32 data)
 {
 	return MASK(capability) & data ? 1 : 0;
@@ -904,6 +1149,8 @@ int capng_have_capability(capng_type_t which, unsigned int capability)
 			return check_inheritable(capability, idx);
 		else if (which == CAPNG_BOUNDING_SET)
 			return bounds_bit_check(capability, idx);
+		else if (which == CAPNG_AMBIENT)
+			return ambient_bit_check(capability, idx);
 	}
 	return 0;
 }
@@ -934,17 +1181,21 @@ char *capng_print_caps_numeric(capng_print_t where, capng_select_t set)
 					m.data.v3[0].permitted,
 					m.data.v3[1].inheritable & UPPER_MASK,
 					m.data.v3[0].inheritable);
-			
 			}
 		}
 #ifdef PR_CAPBSET_DROP
 		if (set & CAPNG_SELECT_BOUNDS)
 			printf("Bounding Set: %08X, %08X\n",
 				m.bounds[1] & UPPER_MASK, m.bounds[0]);
+#endif
+#ifdef PR_CAP_AMBIENT
+		if (set & CAPNG_SELECT_AMBIENT)
+			printf("Ambient :     %08X, %08X\n",
+				m.ambient[1] & UPPER_MASK, m.ambient[0]);
 #endif
 	} else if (where == CAPNG_PRINT_BUFFER) {
 		if (set & CAPNG_SELECT_CAPS) {
-			// Make it big enough for bounding set, too
+			// Make it big enough for bounding & ambient set, too
 			ptr = malloc(160);
 			if (m.cap_ver == 1) {
 				snprintf(ptr, 160,
@@ -970,8 +1221,9 @@ char *capng_print_caps_numeric(capng_print_t where, capng_select_t set)
 		if (set & CAPNG_SELECT_BOUNDS) {
 #ifdef PR_CAPBSET_DROP
 			char *s;
-			if (ptr == NULL ){
-				ptr = malloc(40);
+			// If ptr is NULL, we only room for bounding and ambient
+			if (ptr == NULL ) {
+				ptr = malloc(80);
 				if (ptr == NULL)
 					return ptr;
 				*ptr = 0;
@@ -980,6 +1232,23 @@ char *capng_print_caps_numeric(capng_print_t where, capng_select_t set)
 				s = ptr + strlen(ptr);
 			snprintf(s, 40, "Bounding Set: %08X, %08X\n",
 					m.bounds[1] & UPPER_MASK, m.bounds[0]);
+#endif
+		}
+		if (set & CAPNG_SELECT_AMBIENT) {
+#ifdef PR_CAP_AMBIENT
+			char *s;
+			// If ptr is NULL, we only room for ambient
+			if (ptr == NULL ) {
+				ptr = malloc(40);
+				if (ptr == NULL)
+					return ptr;
+				*ptr = 0;
+				s = ptr;
+			} else
+				s = ptr + strlen(ptr);
+			snprintf(s, 40, "Ambient Set: %08X, %08X\n",
+					m.ambient[1] & UPPER_MASK,
+					m.ambient[0]);
 #endif
 		}
 	}
@@ -1009,7 +1278,7 @@ char *capng_print_caps_text(capng_print_t where, capng_type_t which)
 			} else if (where == CAPNG_PRINT_BUFFER) {
 				int len;
 				if (once == 0) {
-					ptr = malloc(last_cap*18);
+					ptr = malloc(last_cap*20);
 					if (ptr == NULL)
 						return ptr;
 					len = sprintf(ptr+cnt, "%s", n);
