@@ -1,6 +1,6 @@
 /*
  * netcap.c - A program that lists network apps with capabilities
- * Copyright (c) 2009-10, 2012 Red Hat Inc., Durham, North Carolina.
+ * Copyright (c) 2009-10,2012,2020 Red Hat Inc.
  * All Rights Reserved.
  *
  * This software may be freely redistributed and/or modified under the
@@ -15,7 +15,8 @@
  *
  * You should have received a copy of the GNU General Public License
  * along with this program; see the file COPYING. If not, write to the
- * Free Software Foundation, Inc., 675 Mass Ave, Cambridge, MA 02139, USA.
+ * Free Software Foundation, Inc., 51 Franklin Street, Fifth Floor
+ * Boston, MA 02110-1335, USA.
  *
  * Authors:
  *   Steve Grubb <sgrubb@redhat.com>
@@ -59,7 +60,8 @@ static int collect_process_info(void)
 		int pid, ppid;
 		capng_results_t caps;
 		char buf[100];
-		char *tmp, cmd[16], state, *text, *bounds;
+		char *tmp, cmd[16], state;
+		char *text = NULL, *bounds = NULL, *ambient = NULL;
 		int fd, len, euid = -1;
 
 		// Skip non-process dir entries
@@ -94,7 +96,7 @@ static int collect_process_info(void)
 			continue;
 
 		// now get the capabilities
-		capng_clear(CAPNG_SELECT_BOTH);
+		capng_clear(CAPNG_SELECT_ALL);
 		capng_setpid(pid);
 		if (capng_get_caps_process())
 			continue;
@@ -134,18 +136,27 @@ static int collect_process_info(void)
 			fclose(sf);
 		}
 
+		caps = capng_have_capabilities(CAPNG_SELECT_AMBIENT);
+		if (caps > CAPNG_NONE)
+			ambient = strdup("@");
+		else
+			ambient = strdup("");
+		if (!ambient) {
+			fprintf(stderr, "Out of memory\n");
+			free(text);
+			continue;
+		}
+
 		// Now record the bounding set information
-		if (caps == CAPNG_PARTIAL) {
-			caps = capng_have_capabilities(CAPNG_SELECT_BOUNDS);
-			if (caps == CAPNG_FULL)
-				bounds = strdup("+");
-			else
-				bounds = strdup("");
-		} else
+		caps = capng_have_capabilities(CAPNG_SELECT_BOUNDS);
+		if (caps > CAPNG_NONE)
+			bounds = strdup("+");
+		else
 			bounds = strdup("");
 		if (!bounds) {
 			fprintf(stderr, "Out of memory\n");
 			free(text);
+			free(ambient);
 			continue;
 		}
 
@@ -165,6 +176,7 @@ static int collect_process_info(void)
 					strerror(errno));
 			free(text);
 			free(bounds);
+			free(ambient);
 			continue;
 		}
 		// For each file in the fd dir...
@@ -180,7 +192,7 @@ static int collect_process_info(void)
 			if ((llen = readlink(ln, line, sizeof(line)-1)) < 0)
 				continue;
 			line[llen] = 0;
-			
+
 			// Only look at the socket entries
 			if (memcmp(line, "socket:", 7) == 0) {
 				// Type 1 sockets
@@ -208,32 +220,36 @@ static int collect_process_info(void)
 			node.inode = inode;
 			node.capabilities = strdup(text);
 			node.bounds = strdup(bounds);
-			if (node.cmd && node.capabilities && node.bounds)
+			node.ambient = strdup(ambient);
+			if (node.cmd && node.capabilities && node.bounds &&
+			    node.ambient)
 				// We make one entry for each socket inode
 				list_append(&l, &node);
 			else {
 				free(node.cmd);
 				free(node.capabilities);
 				free(node.bounds);
+				free(node.ambient);
 			}
 		}
 		closedir(f);
 		free(text);
 		free(bounds);
+		free(ambient);
 	}
 	closedir(d);
 	return 0;
 }
 
-static void report_finding(int port, const char *type, const char *ifc)
+static void report_finding(unsigned int port, const char *type, const char *ifc)
 {
 	struct passwd *p;
 	lnode *n = list_get_cur(&l);
-		
+
 	// And print out anything with capabilities
 	if (header == 0) {
-		printf("%-5s %-5s %-10s %-16s %-4s %-6s %s\n",
-			"ppid", "pid", "acct", "command", "type", "port", 
+		printf("%-5s %-5s %-10s %-16s %-8s %-6s %s\n",
+			"ppid", "pid", "acct", "command", "type", "port",
 			"capabilities");
 			header = 1;
 	}
@@ -253,12 +269,12 @@ static void report_finding(int port, const char *type, const char *ifc)
 		printf("%-5d %-5d %-10s", n->ppid, n->pid, tacct);
 	} else
 		printf("%-5d %-5d %-10d", n->ppid, n->pid, last_uid);
-	printf(" %-16s %-4s", n->cmd, type);
+	printf(" %-16s %-8s", n->cmd, type);
 	if (ifc)
 		printf(" %-6s", ifc);
 	else
-		printf(" %-6d", port);
-	printf(" %s %s\n", n->capabilities, n->bounds);
+		printf(" %-6u", port);
+	printf(" %s %s%s\n", n->capabilities, n->ambient, n->bounds);
 }
 
 static void read_tcp(const char *proc, const char *type)
@@ -267,7 +283,8 @@ static void read_tcp(const char *proc, const char *type)
 	FILE *f;
 	char buf[256];
 	unsigned long rxq, txq, time_len, retr, inode;
-	int local_port, rem_port, d, state, timer_run, uid, timeout;
+	unsigned int local_port, rem_port, state, timer_run;
+	int d, uid, timeout;
 	char rem_addr[128], local_addr[128], more[512];
 
 	f = fopen(proc, "rte");
@@ -286,7 +303,7 @@ static void read_tcp(const char *proc, const char *type)
 		more[0] = 0;
 		sscanf(buf, "%d: %64[0-9A-Fa-f]:%X %64[0-9A-Fa-f]:%X %X "
 			"%lX:%lX %X:%lX %lX %d %d %lu %511s\n",
-			&d, local_addr, &local_port, rem_addr, &rem_port,
+			&d, &local_addr, &local_port, &rem_addr, &rem_port,
 			&state, &txq, &rxq, &timer_run, &time_len, &retr,
 			&uid, &timeout, &inode, more);
 		if (list_find_inode(&l, inode))
@@ -301,7 +318,8 @@ static void read_udp(const char *proc, const char *type)
 	FILE *f;
 	char buf[256];
 	unsigned long rxq, txq, time_len, retr, inode;
-	int local_port, rem_port, d, state, timer_run, uid, timeout;
+	unsigned int local_port, rem_port, state, timer_run;
+	int d, uid, timeout;
 	char rem_addr[128], local_addr[128], more[512];
 
 	f = fopen(proc, "rte");
@@ -320,7 +338,7 @@ static void read_udp(const char *proc, const char *type)
 		more[0] = 0;
 		sscanf(buf, "%d: %64[0-9A-Fa-f]:%X %64[0-9A-Fa-f]:%X %X "
 			"%lX:%lX %X:%lX %lX %d %d %lu %511s\n",
-			&d, local_addr, &local_port, rem_addr, &rem_port,
+			&d, &local_addr, &local_port, &rem_addr, &rem_port,
 			&state, &txq, &rxq, &timer_run, &time_len, &retr,
 			&uid, &timeout, &inode, more);
 		if (list_find_inode(&l, inode))
@@ -335,7 +353,8 @@ static void read_raw(const char *proc, const char *type)
 	FILE *f;
 	char buf[256];
 	unsigned long rxq, txq, time_len, retr, inode;
-	int local_port, rem_port, d, state, timer_run, uid, timeout;
+	unsigned int local_port, rem_port, state, timer_run;
+	int d, uid, timeout;
 	char rem_addr[128], local_addr[128], more[512];
 
 	f = fopen(proc, "rte");
@@ -354,7 +373,7 @@ static void read_raw(const char *proc, const char *type)
 		more[0] = 0;
 		sscanf(buf, "%d: %64[0-9A-Fa-f]:%X %64[0-9A-Fa-f]:%X %X "
 			"%lX:%lX %X:%lX %lX %d %d %lu %511s\n",
-			&d, local_addr, &local_port, rem_addr, &rem_port,
+			&d, &local_addr, &local_port, &rem_addr, &rem_port,
 			&state, &txq, &rxq, &timer_run, &time_len, &retr,
 			&uid, &timeout, &inode, more);
 		if (list_find_inode(&l, inode))
@@ -449,6 +468,8 @@ int main(int argc, char __attribute__((unused)) *argv[])
 	// Next udp sockets...
 	read_udp("/proc/net/udp", "udp");
 	read_udp("/proc/net/udp6", "udp6");
+	read_udp("/proc/net/udplite", "udplite");
+	read_udp("/proc/net/udplite6", "udplite6");
 
 	// Next, raw sockets...
 	read_raw("/proc/net/raw", "raw");
@@ -457,6 +478,8 @@ int main(int argc, char __attribute__((unused)) *argv[])
 	// And last, read packet sockets
 	read_packet();
 
+	// Could also do icmp,netlink,unix
+
 	list_clear(&l);
 	return 0;
 }
